//*************************************************************************************************
//
// JUMA TX500 & TX136 Transmitter Controller
// Juha Niinikoski, OH2NLT 06.07.2008
//
// Microchip MPLAB + C30 compiler
// MCU dsPIC30F6014A
//
// speed tune added, calibration constants changed 24.08.2009
// 136kHz / 500kHz ref osc selection added 31.08.2009
// command clean up, beacon test play, bug fixes 10.09.2009
// Beacon repeat count logic added 14.09.2009
// More morse characters coded 15.09.2009
// Scandi character LCD display, more TX500 protocol commands, Beacon auto start after power failure 16.09.2009
// very slow CW keyer speed, downto 0.1 wpm 20.09.2009
// CW speed user interface setup limited down to 1 wpm 23.09.2009
// Band limits set  SW v1.00, 10.10.2009
// Beacon end quick TX release 11.10.2009, SW v1.01
// Beacon Transmit =BT command added 15.10.2009, SW v1.02
// beacon message '=' and '?' character behaviour corrected 17.10.2009 SW v1.03
// 0.1 wpm CW bug corrected, beep(int, unsigned int), cw_period => unsigned int  24.03.2010
// Add QRSS and DFCW modes, SW 1.05a - F4GCB 08.2014
// Add WSPR mode, SW 1.06a - F4GCB 10.2014
// Bug fixed: add init WSPR beacon for automatically restart, SW 1.06b - F4GCB 12.2014
// Add declarations to avoid compilation warnings - F4GCB 12-2014
// Program structure modified to improve readability and maintenance - F4GCB 02-2015
// Add setting configuration values - F4GCB 02-2015
// Add OPERA mode - F4GCB 05.2015
// Add JASON mode, SW 1.07 - F4GCB 08.2015
// Add CW identity option - F4GCB 04.2016
// Add SCRIPT mode - F4GCB 04.2016
// Limit EEPROM writing to preserve byte endurance, SW 1.08 - F4GCB 05.2016
//
//*************************************************************************************************

#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "juma-tx500.h"						// JUMA-TRX2 hardware specific definitions
#include "eeprom_tx500.h"
#include "lcd_tx500.h"						// LCD specific definitions
#include "serial_test.h"
#include "serial_tx500.h"
#include "ad9833.h"
#include "gene_jason.h"
#include "gene_morse.h"
#include "gene_opera.h"
#include "gene_wspr.h"
#include "timers_pwm.h"
#include "uart.h"
#include "uart2.h"

// Definitions ------------------------------------------------------------------------------------
#define MAX_LCD_MODE 1						// last display "main" page			/
#define	MAX_SUB_PAGE0 4						// sub page 0 displays				|
#define	MAX_SUB_PAGE1 12					// sub page 1 displays				|> display control
#define	MAX_SUB_PAGE2 7						// sub page 2 displays				|
#define S_BAR 0x07							// S-meter vertical bar character	/
#define MAX_SERVICE_PAGES 8					// number of display pages for service mode


// ????? ------------------------------------------------------------------------------------------ 
_FOSC(CSW_FSCM_OFF & XT_PLL4);				// 7,3728MHz XT, 29,4912MHz clock = 7,3728MHz cycle
//_FOSC(CSW_FSCM_OFF & XT_PLL8);			// 3,68640MHz XT, 29,4912MHz clock = 7,3728MHz cycle
_FWDT(WDT_OFF);                				//Turn off the Watch-Dog Timer.
_FBORPOR(MCLR_EN & PWRT_64 & BORV_20);   	//Enable MCLR, power on timer, brown out 2V
//_FBORPOR(MCLR_EN & PWRT_64 & BORV_27);   	//Enable MCLR, power on timer, brown out 2V
//_FGS(GEN_PROT);            				//Disable Code Protection, old compiler
_FBS(CODE_PROT_OFF);
_FSS(CODE_PROT_OFF);
_FGS(CODE_PROT_OFF);


// Constant variables -----------------------------------------------------------------------------
const char ver[] = {"1.08"};			// software version information, displayed at startup
const char date[] = {"11.05.2016"};		// version date


// Global variables -------------------------------------------------------------------------------
int board_type;							// main board type, 1=TX500, 2=TX136, 0=unknown
unsigned int alarms;					// alarm bits
int cw_active;							// cw active flag for keyer code
int keyer_mode;							// keyer mode selector
int key;								// TX request status


// Local variables --------------------------------------------------------------------------------
static int fcy;							// clock / kHz
static int service_mode;				// service mode selected flag
static int alarm_beep;					// alarm beep logic flag

static int lcd_mode;					// lcd display mode 0=Display 1=Config 2=Mode setup
static int sub_page0;					// lcd display sub page for lcd_mode0
static int sub_mode;					// lcd display sub mode for lcd_mode0
static int sub_mode_flg = 0;			// lcd display sub mode flag
static int sub_page1;					// lcd display sub page for lcd_mode1
static int sub_page2;					// lcd display sub page for lcd_mode2

static int batt_samples = 0;			// battery voltage meter
static unsigned long batt_mean = 0;
static long out_pwr;					// scaled output power = fwd_pwr - rev_pwr * scale
static long peak_pwr;					// peak, fast attach slow decay power
static int swr;							// calculated swr * 100
static int old_swr;						// for alarm trigger
static int pwr_meter;					// scaled S-meter value 0 ... 48

static char beacon_lcd[260];			// beacon message for display
static int beacon_lcd_idx;				// beacon message display index
static int beacon_lcd_roll;				// speed flag for rolling beacon message display

static const char br_txt[8][8] = {{"1200  "}, {"2400  "}, {"4800  "} ,{"9600  "}, {"19200 "}, {"38400 "}, {"57600 "}, {"115200 "}}; // Baud rate


// Local functions --------------------------------------------------------------------------------
static void service(void);					// Service & calibration functions, when tested move to own module
static void disp_timer(int);				// Display WSPR timer


// External functions -----------------------------------------------------------------------------
extern void us_delay( unsigned int);		// delay routines
extern void ms_delay( unsigned int);

extern void init_adc12(void);				// ADC routines
extern int convert_adc12(unsigned int);



//*************************************************************************************************
// On/Off, Alarms and RS232 test functions
//*************************************************************************************************

//=================================================================================================
// Handle PWR button. Shut down the system
//=================================================================================================
void power_off(void)
{
	const char line1[] = {"Pwr OFF after %ld "};
	const char line2[] = {"                "};
	char buffer[17];

	beep(HZ466_85, cal.calval.beep_len);								// notify PWR button pressed
	ms_delay(250);

	lcd_clear();
	counter.push = POWER_OFF;											// load power off guard timer

	while (PWR_SW == 1)
	{
		sprintf(buffer, line1, (counter.push/300));						// display off counter, 3,2,1,0
		lcd_display_all(buffer, line2);

		if (counter.push == 0)
		{
			beep(HZ392_01, cal.calval.beep_len*2);						// play low sound when ready for power down
			while (PWR_SW == 1) {};
		}
	}

	if (counter.push != 0)												// too short push, continue normal operation
		return;

	save_defaults();
	//save_beacon();
	//save_script();
 	if (eeprom.defval.serial_mode == 1)	// conditional print only if RS232 = Test
		printf("\n\rUser settings saved to EEPROM, Power off\n\r");

	ms_delay(2);														// let the printout finish
	PWR_ON = 0;
	for (;;) {};
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Check if any alarms
//=================================================================================================
void check_alarms(void)
{
	// check over current alarm bit & set alarm
	if (OC)
	{
		CW = 0;									// RF off immediately
		alarms = alarms | CURR_AL;				// set alarm flag
	}

	// check SWR, two samples neded for trip
	// Note, SWR alarm is active also in STBY
	if (swr > eeprom.defval.swr_limit[eeprom.defval.rfpwr] && old_swr > eeprom.defval.swr_limit[eeprom.defval.rfpwr])
	{
		CW = 0;									// RF off immediately
		alarms = alarms | SWR_AL;				// set alarm flag
	}

	if (alarms != 0)
	{
		lcd_mode = 0;							// force normal display
		set_pa_state(0, 1, 0, 0);				// force transmitter to STBY
		set_beacon_on_off(0, 0);				// force beacon Off
	}
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Clear alarms & try to reset over current latch
//=================================================================================================
void clear_alarms(void)
{
	alarms = 0;									// clear alarm flags
	alarm_beep = 0;

	OC_CLR = 0;									// generate reset pulse for over current latch
	ms_delay(20);
	OC_CLR = 1;
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// RS232 loop back test
//=================================================================================================
void rs232_test(void)
{
	unsigned char c;
	const char line1[] = {"RS232 Test      "};
	const char line2[] = {"Waiting data...."};
	const char data[] = {"RX data  %c  0x%x"};
	char buffer[17];

	lcd_clear();
	lcd_display_all(line1, line2);

	for (;;)
	{
		if (uart1_kbhit())
		{
			c = uart1_get_char();
			beep(HZ698_45, 10);				// give beep tone
			uart1_put_char(c);				// echo character
			sprintf(buffer, data, c, c);
			lcd_display_line(2, buffer);
		}

		if (PWR_SW == 1)						// exit test
			break;
	}
}
//-------------------------------------------------------------------------------------------------


//*************************************************************************************************
// Calculate and measure values
//*************************************************************************************************

//=================================================================================================
// Check I/O board type
//=================================================================================================
int check_io_board_type(void)
{
	int x, b;

	x = convert_adc12(B_TYPE);				// get board ID voltage
	b = UNKNOWN_TYPE;						// board identified

	if (x > TX500_LOW && x < TX500_HI)		// is it TX500
		b = TX500_TYPE;

	if (x > TX136_LOW && x < TX136_HI)		// is it TX136
		b = TX136_TYPE;

	return b;
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Calculate SWR
//=================================================================================================
int calc_swr(int fwd, int rev)
{
	int swr;						// swr calculated
	long Ro100;						// 100*Ro
	long x;							// 100*(1-Ro)

	old_swr = swr;

	//  rev/fwd ratio * 100
	Ro100 = ((long)rev * 100L) / fwd;
	x = 100 - Ro100;

	// prevent divide by zero
	if (x < 1)
		x = 1;

	// calculate only if we have some signal
	if (fwd >= 50)
		// 100 * (1+Ro)/(1-Ro)
		swr = (((100 + Ro100)*100) / x);
	else
		// no swr reading if low power
		swr = 0;

	// max display value 99.99
	if (swr > 9999)
		swr = 9999;

	// value 100 = swr 1.00
	return swr;
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Power measurements
//=================================================================================================
void power_measurements(void)
{
	int fwd_adval, rev_adval;
	long fwd_pwr, rev_pwr;

	// convert fwd & rev A/D channells
	fwd_adval = convert_adc12(FWD_PWR);
	rev_adval = convert_adc12(REV_PWR);

	// store not scaled values for later use
	// calculate relative power levels (U^2)
	fwd_pwr = (long)fwd_adval * (long)fwd_adval;
	rev_pwr = (long)rev_adval * (long)rev_adval;

	// output power = forward - reverse.
	out_pwr = fwd_pwr - rev_pwr;

	// forward voltage display
	pwr_meter = (fwd_adval * F_VOLT_GRAPH_MULT) >> 17;

	// shift sample queue and calculate SWR
	old_swr = swr;
	swr = calc_swr(fwd_adval, rev_adval);
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Get output power
//=================================================================================================
int get_out_pwr(void)
{
	return (out_pwr * cal.calval.fwd_pwr_mult) >> 16;
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Get SWR
//=================================================================================================
int get_swr(void)
{
	return swr;
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Measure battery voltage
//=================================================================================================
unsigned long meas_batt_volt(void)
{
	return ((unsigned long)convert_adc12(BATT_CH) * cal.calval.batt_mult) >> 8;
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Measure drain current
//=================================================================================================
long meas_id(void)
{
	return ((long)convert_adc12(ID_CUR) * cal.calval.id_mult) >> 16;
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// decrease dBm level in function TX power for WSPR mode
//=================================================================================================
int ndbm_rfpwr(void)
{
	int nu_rfpwr[4] = {13,6,3,0};

	return wspr_ndbm(eeprom.defval.wspr_ndbm - nu_rfpwr[eeprom.defval.rfpwr]);
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Get TX500/136 info
//=================================================================================================
char *get_info(char *info)
{
	char type[3][11] = {
		{"UNKNOWN"},
		{"JUMA-TX500"},
		{"JUMA-TX136"}};

	sprintf(info, "%s, SW %s, DATE %s", type[board_type], ver, date);
	return info;
}
//-------------------------------------------------------------------------------------------------


//*************************************************************************************************
// Beacon management
//*************************************************************************************************

//=================================================================================================
// Set WSPR timer
//=================================================================================================
void set_wspr_timer(int value, int new_val, int gps, int irq, int beep_on, int display_on)
{
	const char line1[17] = {"WSPR Timer      "};

	// set WSPR timer value by GPS
	if (gps)
	{
        counter.wspr = value;

		// set timers_pwm and gps synchronization counters
		counter.ms = 1000 - irq;
		counter.ms_offset = 0;
		counter.gps = 3;
	}

	// set WSPR timer value by IRQ
    else if (irq)
        counter.wspr = counter.wspr + value;

	// set WSPR timer value
	else if (new_val)
	{
		counter.wspr = value;

		// set timers_pwm counters
        counter.ms = 1000;
		counter.ms_offset = 0;
	}
		
	// set WSPR timer value by button (set minutes only)
	else if (value && counter.gps == 0)
	{
		counter.wspr = ((counter.wspr / 60) + value) * 60;

		// set timers_pwm counters
        counter.ms = 1000;
		counter.ms_offset = 0;
	}

	if (value && !gps)
	{
        // check value
        if (counter.wspr >= 3600)
			counter.wspr = 0;
		else if (counter.wspr <= -1)
			counter.wspr = 3540;

        // sound tone for each second (help to set manually)
        if (beep_on && !counter.gps && lcd_mode == 1 && sub_page1 == 1)
			beep_up_down(&value);
	}

	if (display_on)
	{
		lcd_display_line(1, line1);
		disp_timer(1);
	}
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Init beacon
//=================================================================================================
void init_beacon(void)
{
	int cwid_speed[3] = {0, 12, 24};
	char buffer[38];
	char cmd_cwid[9];
	char cmd_cw[4];
	char cmd_mode[4];

	// define commands and place escape sequences for CW identity
	if (eeprom.defval.cwid)
	{
		sprintf(cmd_cwid, "%cg0%cs%2d", 0x5C, 0x5C, cwid_speed[eeprom.defval.cwid]);
		sprintf(cmd_cw, "%cs%2d", 0x5C, eeprom.defval.cw_speed);
		sprintf(cmd_mode, "%cg%1d", 0x5C, eeprom.defval.mode);
	}
 
	switch (eeprom.defval.mode)
	{
		case MODE_CW: default:
			morse_init_beacon(beacon_buffer);
		break;

		case MODE_QRSS: case MODE_DFCW:
			if (eeprom.defval.cwid)					// CW identity available
			{
				sprintf(buffer, "%s%s %s %s%s", eeprom.defval.short_msg, cmd_cwid, eeprom.defval.wspr_call, cmd_cw, cmd_mode); 
				flag.cwid = 1;
			}
			else
				sprintf(buffer, eeprom.defval.short_msg);
			morse_init_beacon(buffer);
		break;

		case MODE_JASON:
			jason_init_beacon(eeprom.defval.short_msg);
		break;

		case MODE_OPERA:
			opera_init_beacon(eeprom.defval.wspr_call);
			counter.s = 0;
		break;

		case MODE_WSPR:
			if (eeprom.defval.wspr_gps == 1 && counter.gps)
				wspr_init_beacon(eeprom.defval.wspr_call, get_gps_grid4(buffer), ndbm_rfpwr());
			else
				wspr_init_beacon(eeprom.defval.wspr_call, eeprom.defval.wspr_grid, ndbm_rfpwr());
		break;
	}

	// CW identity available
	if (eeprom.defval.cwid && eeprom.defval.mode > MODE_DFCW)
	{
		sprintf(buffer, "%s %s %s%s", cmd_cwid, eeprom.defval.wspr_call, cmd_cw, cmd_mode);
		morse_init_beacon(buffer);
		flag.cwid = 1;
	}
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Set beacon on/off
//=================================================================================================
void set_beacon_on_off(int tx_on, int beep_on)
{
	const int tone = TONE_B;

	// beacon was off
	if(tx_on && eeprom.defval.beacon_tx_on == 0)
	{
		init_beacon();
		flag.beacon = 1;
		old.mode = eeprom.defval.mode;
		old.pa_state = eeprom.defval.pa_state;
		old.keyer = keyer_mode;

		switch (eeprom.defval.mode)
		{
			case MODE_CW: case MODE_QRSS: case MODE_DFCW: default:
				if (eeprom.defval.script_on)					// one play if script
					set_beacon(1, 1);
				else if (eeprom.defval.cw_frame == 0)
					set_beacon(-1, 1);							// continuous play
				else
					set_beacon(eeprom.defval.cw_frame, 1);		// rounds left(positive)
				
			break;

			case MODE_JASON:
				if (eeprom.defval.script_on)					// one play if script
					set_beacon(1, 1);
				else if (eeprom.defval.jason_frame == 0)
					set_beacon(-1, 1);							// continuous play
				else
					set_beacon(eeprom.defval.jason_frame, 1);	// rounds left(positive)
			break;

			case MODE_OPERA:
				if (eeprom.defval.opera_frame == 0 || eeprom.defval.script_on)
					set_beacon(1, 1);							// one play or one play if script
				else
					set_beacon(-1, 1);							// continuous play according to the frame
			break;

			case MODE_WSPR:
				if (eeprom.defval.wspr_frame == 0 || eeprom.defval.script_on)
					set_beacon(1, 1);							// one play or one play if script
				else
					set_beacon(-1, 1);							// continuous play according to the frame
			break;
		}

		set_pa_state(1, 1, 0, 0);							// force OPER
		set_cw_keyer(4 , 1, 0, 0);							// force CW keyer mode = Beacon
		if (cal.calval.power_q4 && !eeprom.defval.script_on)
			save_defaults();								// save system state, beacon restart automatically after power failure
		set.cw_break = cal.calval.cw_break_time * 4;		// extended cw break time

		if (eeprom.defval.mode == MODE_DFCW)				// display beacon in progress
			sub_mode = 4;
		else
			sub_mode = 3;
	}

	// beacon was on
	else
	{
		set_beacon(0, 1);										// turn it off
		stop_keyer();
		counter.cw_break = 10;									// fast release 10 ms
		set.cw_break = cal.calval.cw_break_time;				// normal cw break time
		encoder_get();											// clear encoder
		sub_mode = 0;
		if (flag.cwid)
		{
			set_mode(old.mode, 1, 0, 0);						// if beacon stopped during cwid tx
			flag.cwid = 0;
		}
	}

	if (beep_on && !eeprom.defval.script_on)
		beep(tone, cal.calval.beep_len*10);						// long beep for Beacon mode change
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Run beacon
//=================================================================================================
void run_beacon(void)
{
	switch (eeprom.defval.mode)
	{
		case MODE_CW: case MODE_QRSS: case MODE_DFCW: default:
			if (morse_run_beacon())
				set_beacon(-1,0);
			if (eeprom.defval.beacon_tx_on == 0)
				set_beacon_on_off(0,1);
		break;

		case MODE_JASON:
			if (jason_run_beacon())
			{
				if (flag.cwid)
					set_mode(MODE_CW, 1, 0, 0);
				else
					set_beacon(-1,0);
			}
			if (eeprom.defval.beacon_tx_on == 0)
				set_beacon_on_off(0,1);
		break;

		case MODE_OPERA:
			if (opera_run_beacon(eeprom.defval.opera_speed, eeprom.defval.opera_frame, &counter.s))
			{
				if (flag.cwid)
					set_mode(MODE_CW, 1, 0, 0);
				else
					set_beacon(-1,0);
			}
			if (eeprom.defval.beacon_tx_on == 0)
				set_beacon_on_off(0,1);
		break;

		case MODE_WSPR:
			if (wspr_run_beacon(eeprom.defval.wspr_speed, eeprom.defval.wspr_frame, counter.wspr))
			{
				if (flag.cwid)
					set_mode(MODE_CW, 1, 0, 0);
				else
					set_beacon(-1,0);
			}
			if (eeprom.defval.beacon_tx_on == 0)
				set_beacon_on_off(0,1);
		break;
	}
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Set script on/off
//=================================================================================================
void set_script_on_off(int on, int beep_on)
{
	const int tone = TONE_B;

	// script was off
	if (on)
	{
		serial_script_init(script_buffer);
		if (eeprom.defval.script_frame == 0)
			set_script_on(-1, 1);							// continuous play
		else
			set_script_on(eeprom.defval.script_frame, 1);	// rounds left(positive)

		if (cal.calval.power_q4)
			save_defaults();								// save system state, beacon restart automatically after power failure
	}

	// script was on
	else
	{
		set_script_on(0, 1);
		counter.script = 0;
		sub_mode = 0;
		set_mode(MODE_SCRIPT, 1, 0, 0);
	}

	if (beep_on)
		beep(tone, cal.calval.beep_len*10);				// long beep for script mode change
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Run script
//=================================================================================================
void run_script(void)
{	
	// if beacon in progress, no script reading
	if (eeprom.defval.beacon_tx_on == 0)
	{
		// script time delay
		if (counter.script)
		{
			if (!sub_mode_flg)
			{
				sub_mode = 2;
				sub_mode_flg = 1;
			}
			return;
		}
		else if (sub_mode_flg)
			sub_mode_flg = 0;

		// script reading
		if (serial_script_run())
			set_script_on(-1, 0);
		if (eeprom.defval.script_on == 0)
			set_script_on_off(0,1);
	}
}
//-------------------------------------------------------------------------------------------------


//*************************************************************************************************
// Display functions
//*************************************************************************************************

//=================================================================================================
// Display hello message
//=================================================================================================
void disp_hello(void)
{
	const char line1[3][14] = {
		{"JUMA ????? %s"},
		{"JUMA TX500 %s"},
		{"JUMA TX136 %s"}};
	const char line2[] = {"  OH2NLT OH7SV  "};
	char buffer[17];

	lcd_clear();
	sprintf(buffer, line1[board_type], ver);
	lcd_display_all(buffer, line2);
}
//-------------------------------------------------------------------------------------------------


//*************************************************************************************************
// Display meter functions
//*************************************************************************************************

//=================================================================================================
// Display slarms
//=================================================================================================
void disp_alarms(void)
{
	if(flag.blink == 0)
	{
		lcd_display_at(2, 5, "      ");			// show nothing
		alarm_beep = 0;
	}
	else										// show alarm, only highest priority is shown
	{
		if(alarm_beep == 0)						// trigger with rising edge
		{
			alarm_beep = 1;
			beep(HZ698_45, ALARM_BEEP_LEN);		// make sound
		}
		if(alarms & SWR_AL)
			lcd_display_at(2, 5, " SWR  ");
		else if(alarms & CURR_AL)
			lcd_display_at(2, 5, " CURR ");
		else if (alarms & MSG_AL)
			lcd_display_at(2, 5, " MSG  ");
		else
			lcd_display_at(2, 5, " ???? ");		// unknown
		lcd_cursor_on(2, 5, 0);					// cursor off if alarm in beacon configuration mode
	}
}
//-------------------------------------------------------------------------------------------------

//=================================================================================================
// Display watt meter
//=================================================================================================
void disp_fwd_pwr(void)
{
	long decay_term;
	double temp;
	const char ok_pwr[] = {"P %4.1fW"};
	char buffer[9];

	if(key)													// TXon ?
	{
		if(out_pwr > peak_pwr)								// fast attach
		{
			peak_pwr = out_pwr;								// fast attach
			counter.decay = PEAK_SHOW_TIME;					// reload decay timer
		}
		else												// slow release
		{
			if(counter.decay == 0)
			{
				counter.decay = DECAY_TIME;					// reload decay timer
				decay_term = peak_pwr >> DECAY_RATE;

				if(decay_term == 0)							// zero reached
					peak_pwr = 0;
				else
					peak_pwr = peak_pwr - decay_term;		// decay
			}
		}
		temp = (peak_pwr * cal.calval.fwd_pwr_mult) >> 16;	// scaling, P*100 = ADC^2 * scaling / 65536
		sprintf(buffer, ok_pwr, temp / 10);
		lcd_display_at(1, 10, buffer);
	}
	else
	{
		set_tx_power(0, 0, 0, 1);							// display setting
		peak_pwr = 0;
	}
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Display SWR meter
//=================================================================================================
void disp_swr(void)
{
	const char no_swr[] = {"SWR N/A"};
	const char ok_swr[] = {"SWR%4.1f"};
	char buffer[9];

	if(key)								// TXon ?
	{
		sprintf(buffer, ok_swr, (double)swr / 100);
		lcd_display_at(1, 10, buffer);
	}
	else
		lcd_display_at(1, 10, no_swr);
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Display battery voltage meter
//=================================================================================================
void disp_batt_volt(void)
{
	const char batt[] = {" %5.2fV"};
	char buffer[9];

	if(batt_samples > 15)									// 16 samples done ?
	{
		sprintf(buffer, batt, (double)batt_mean / 1600);
		lcd_display_at(1, 10, buffer);
		batt_samples = 0;
		batt_mean = 0;
	}
	else
	{
		batt_mean += meas_batt_volt();						// measure & add
		batt_samples++;										// add sample counter
	}
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Display Drain current meter
//=================================================================================================
void disp_id(void)
{
	const char no_id_cur[] = {"I --.-A"};
	const char ok_id_cur[] = {"I %4.1fA"};
	char buffer[9];

	if(key)								// TXon ?
	{
		sprintf(buffer, ok_id_cur, (double)meas_id() / 10);
		lcd_display_at(1, 10, buffer);
	}
	else
		lcd_display_at(1, 10, no_id_cur);
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Display WSPR timer
//=================================================================================================
void disp_timer(int area)
{
	const char no_gps[] = {"T %.2d:%.2d"};
	const char ok_gps[] = {"T^%.2d:%.2d"};
	char buffer[9];

	if (counter.gps)
		sprintf(buffer, ok_gps, (counter.wspr / 60), (counter.wspr % 60));
	else
		sprintf(buffer, no_gps, (counter.wspr / 60), (counter.wspr % 60));

	if (area)
	{
		lcd_display_line(2, buffer);
		lcd_display_at(2, 8, "         ");
	}
	else
		lcd_display_at(1, 10, buffer);
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Display beacon message in progress
//=================================================================================================
void disp_beacon_in_progress(void)
{
	char buffer[10];

	if (eeprom.defval.beacon_tx_on == 0)
	{	
		sub_mode = 0;
		return;
	}

	switch (eeprom.defval.mode)
	{
		case MODE_CW: case MODE_QRSS: case MODE_DFCW: default:
			lcd_display_at(1, 1, morse_get_display(buffer));
		break;

		case MODE_JASON:
			lcd_display_at(1, 1, jason_get_display(buffer));
		break;

		case MODE_OPERA:
			lcd_display_at(1, 1, opera_get_code(buffer));
		break;

		case MODE_WSPR:
			lcd_display_at(1, 1, wspr_get_code(buffer));
		break;
	}
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Display beacon message to transmit
//=================================================================================================
void disp_beacon_lcd(int display_mode)
{
	int i;
	char buffer[9];

	if (beacon_lcd_roll != flag.roll)					// speed for rolling display
	{
		switch (eeprom.defval.mode)
		{
			case MODE_CW: default:
				if (flag.cwid)
					sprintf(beacon_lcd, "    Id [%s]       ", eeprom.defval.wspr_call);
				else
					sprintf(beacon_lcd, "    Msg [%s]       ", beacon_buffer);
			break;

			case MODE_QRSS: case MODE_DFCW: case MODE_JASON:
				if (eeprom.defval.cwid)
					sprintf(beacon_lcd, "    Msg [%s]+Id       ", eeprom.defval.short_msg);
				else
					sprintf(beacon_lcd, "    Msg [%s]       ", eeprom.defval.short_msg);
			break;

			case MODE_OPERA:
				if (eeprom.defval.cwid)
					sprintf(beacon_lcd, "    Msg [%s]+Id       ", eeprom.defval.wspr_call);
				else
					sprintf(beacon_lcd, "    Msg [%s]       ", eeprom.defval.wspr_call);
			break;

			case MODE_WSPR:
				if (eeprom.defval.wspr_gps == 1 && counter.gps)
				{
					if (eeprom.defval.cwid)
						sprintf(beacon_lcd, "    Msg [%s %s %d]+Id       ", eeprom.defval.wspr_call, get_gps_grid4(buffer), ndbm_rfpwr());
					else
						sprintf(beacon_lcd, "    Msg [%s %s %d]       ", eeprom.defval.wspr_call, get_gps_grid4(buffer), ndbm_rfpwr());
				}
				else
				{
					if (eeprom.defval.cwid)
						sprintf(beacon_lcd, "    Msg [%s %s %d]+Id       ", eeprom.defval.wspr_call, eeprom.defval.wspr_grid, ndbm_rfpwr());
					else
						sprintf(beacon_lcd, "    Msg [%s %s %d]       ", eeprom.defval.wspr_call, eeprom.defval.wspr_grid, ndbm_rfpwr());
				}
			break;

			case MODE_SCRIPT:
				sprintf(beacon_lcd, "    Script [%s]       ", script_buffer);
			break;
		}

		beacon_lcd_roll = flag.roll;
		for (i=0; i<8; i++)
			buffer[i] = beacon_lcd[i+beacon_lcd_idx];
		buffer[8] = ' ';
		lcd_display_at(1, 1, buffer);

		if (beacon_lcd_idx < strlen(beacon_lcd) - 8)
			beacon_lcd_idx++;
		else
			beacon_lcd_idx = 0;
	}
}
//-------------------------------------------------------------------------------------------------

//=================================================================================================
// main
//=================================================================================================
int main(void)
{

//*************************************************************************************************
// ----------> Main first init part
//*************************************************************************************************

	int temp;
	unsigned int y, w;

	// set general I/O
	TRISA = INIT_TRISA;				// SWITCHES, PTT
	PORTA = INIT_PORTA;

	TRISB = INIT_TRISB;				// ADC inputs, TXEN, RXEN, General I/O
	PORTB = INIT_PORTB;

	TRISC = INIT_TRISC;				// sideband select & switches
	PORTC = INIT_PORTC;

	TRISD = INIT_TRISD;				// LEDs & LCD pins = output
	PORTD = INIT_PORTD;

	TRISF = INIT_TRISF;				// DDS controls
	PORTF = INIT_PORTF;

	TRISG = INIT_TRISG;				// DDS controls, EEPROM, DScard signals
	PORTG = INIT_PORTG;

	// Power switch
	PWR_ON = 1;								// secure power ON

	// start PWM system
	set.cw_period = 100;					// set something before timer system starts
	//IPC0 =0X5444;							// set tone generator (TMR1) priority higher than others
	init_timers_pwm();						// setup pwm & tone generators
	fcy = FCY / 1000;						// clock freq (kHz)

	// start LCD
	lcd_init();								// init LCD
	lcd_set_chgen();						// load bar graph fonts

	// start UART
	uart1_init();							// wake up serial interfaces
	uart2_init();

	// start ADC
	init_adc12();							// setup A/D converter, must be done before Keyer is started, RB I/O bits

	// check I/O boards
	board_type = check_io_board_type();		// measure & set main board type

	// put hello messages to LCD
	disp_hello();

	// Read calibration & default values from EEPROM
	y = read_calval();
	w = read_defaults();
 	if (eeprom.defval.serial_mode == 1)		// conditional print only if RS232 = Test
		printf("\n\rEEPROM checksums, Cal = %X, Def Cs = %X, Factory default resets = %d\n\r",y, w, fd_counter);
	if (y != 0 || w != 0)					// compare checksums
	{
		if(eeprom.defval.serial_mode == 1)	// conditional print only if RS232 = Test
		{
			dump_eeprom();					// test, show EEPROM content before reset
			printf("\n\rEEPROM checksum error, reload factory defaults");
		}

		set_factory_defaults();				// EEPROM empty or corrupted, set factory defaults
		save_calval();
		save_defaults();
	}

	// set LCD contrast & back light with EEPROM data
	set_pwm4_dac(eeprom.defval.contrast);
	set_pwm3_dac(eeprom.defval.back_light);

	// set display mode etc
	lcd_mode = 0;									// select default display
	sub_page0 = 0;
	sub_mode = 0;
	sub_page1 = 0;
	sub_page2 = 0;
	beacon_lcd_idx = 0;

	// start up beep
	beep(HZ698_45, cal.calval.beep_len * 2);		// give start up beep

	OC_CLR = 0;										// clear over current latch
	ms_delay(1);
	OC_CLR = 1;

	// set user baud rate
	ms_delay(20);									// wait TX empty
	uart1_set_baud(eeprom.defval.br);

	// CW identification flag for CW mode
	flag.cwid = 0;

	// conditional start up text printout, print only if RS232 = Test
 	if(eeprom.defval.serial_mode == 1)
	{
		switch(board_type)							// select hello message by board type
		{
			case TX500_TYPE:
				printf("\n\r\n\rJUMA TX500 / ");
			break;

			case TX136_TYPE:
				printf("\n\r\n\rJUMA TX136 / ");
			break;

			case UNKNOWN_TYPE: default:
				printf("\n\r\n\rJUMA Unknown Main Board type/ ");
			break;
		}
		printf(ver);
		printf(" / ");
		printf(date);
		printf("\n\rCopyright Juha Niinikoski, OH2NLT\n\r\n\r");
		printf("System Clock = %i kHz\n\r", fcy);

		// serial baud rate info
		if(eeprom.defval.br != DEFAULT_BAUD_RATE)
			printf("\n\rSet UART user baud rate to #%i,  %sbd\n\r", eeprom.defval.br, br_txt[eeprom.defval.br]);	// print with user baud rate

		printf("\n\rCW break time set to %i units", cal.calval.cw_break_time);	// show CW break time setup
	}


//*********************************************************************************************
// ----------> Do front panel switches during startup
//*********************************************************************************************

	//-----------------------------------------------------------------------------------------
	// POWER swith long push (service mode)
	//-----------------------------------------------------------------------------------------
	service_mode = 0;
	counter.push = VERY_LONG_PUSH;									// set long push timer
	while(PWR_SW == 1)
	{
		if(counter.push == 0)
		{
			beep(HZ392_01, 300);									// give service beep
			lcd_display_line(2, "Service Mode    ");
			ms_delay(1000);
			service_mode = 1;

			while(PWR_SW == 1) 										// show message as login as Power switch is pressed
				lcd_display_all("Push DISPLAY nxt", "Push OPER = Save");// display some instructions
		}
	}
	if(service_mode != 0)											// do calibration functions if service mode start
		service();

	//-----------------------------------------------------------------------------------------
	// OPER swith long push (factory defaults)
	//-----------------------------------------------------------------------------------------
	while(OPER == 0)								// if OPER button pressed ask user what to do
	{
		lcd_display_all("Push RF PWR for ", "factory defaults");
		if(RFPWR == 0)								// factory defaults asked
		{
			set_factory_defaults();					// EEPROM empty, set factory defaults
			save_calval();
			save_defaults();
			break;
		}
	}

	//-----------------------------------------------------------------------------------------
	// DISP swith long push (RS232 I/O test)
	//-----------------------------------------------------------------------------------------
	if(DISP == 0)						// If Config/DISPLAY pressed during startup
		rs232_test();					// goto RS232 test loop


//*************************************************************************************************
// ----------> Main second init part after possible new calibration values setting
//*************************************************************************************************

	// main board relays
	set_tx_power(0, 0, 0, 0);					// RF power selector
	set_preamp(0, 0, 0, 0);						// set pre amplifier
	set_converter(0, 0, 0, 0);					// select up converter
	set_spare_io(0, 0, 0, 0);					// set SPARE digital I/O state

	// other
	PTT_OUT_1 = 0;
	PTT_OUT_2 = 0;

	// start DDS
	calc_hz_bit();								// calculate Hz/bit from ref osc actual value
	calc_tword();
	calc_morse_shift();							// calcultate TX MORSE shift for DDS
	calc_jason_shift();							// calcultate TX JASON shift for DDS
	calc_wspr_shifts();							// calcultate TX WSPR shifts for DDS
	init_ad9833();
	tune_ad9833(0);								// keep DDS stopped for now

	// keyer setup
	set.cw_break = cal.calval.cw_break_time; 	// normal cw break time
	calc_cw_period();							// calculate CW period value(ms)
	calc_sidetone();							// calculate sidetone timer value
	read_beacon();								// read MORSE beacon message from the EEPROM
	read_script();								// read script from the EEPROM
	if (eeprom.defval.beacon_tx_on && cal.calval.power_q4)
	{
		init_beacon();							// init beacon for automatically restart
		if (eeprom.defval.mode == MODE_DFCW)	// display beacon in progress
			sub_mode = 4;
		else
			sub_mode = 3;
	}
	if (eeprom.defval.script_on && cal.calval.power_q4)
		serial_script_init(script_buffer);		// init script for automatically restart

	ms_delay(2000);								// show start up message

	if(board_type == UNKNOWN_TYPE)				// stop here if unknown board
	{
		printf("\n\r\n\rUnknown main board type, press OPER to continue anyway\n\r");
		lcd_display_at(1, 0, "OPER = Continue ");
		while(OPER == 1) {}						// can continue with OPER button
	}

	// start with clean display
	lcd_clear();


//*************************************************************************************************
// ----------> Main forever loop
//*************************************************************************************************
	for(;;)
	{
		MAIN_TEST = !MAIN_TEST;				// timing test

		// Make atomic copy of KEY (TX request) signal for this user interface processing round
		if (PTT_OUT_1)						// 1 = active
			key = 1;						// copy I/O bit to status flag
		else
			key = 0;
	
		// Check alarm conditions
		check_alarms();						// check if any alarms

		// Set keyer active flag
		if(eeprom.defval.pa_state)			// OPER or TUNE
			cw_active = 1;					// cw active flag for keyer code
		else								// Standby
			cw_active = 0;					// no TX allowed

		// set keyer mode
		if(eeprom.defval.pa_state == STATE_TUNE)		// if TUNE selected
			keyer_mode = KEYER_STRAIGHT;				// force straight
		else if(eeprom.defval.mode == MODE_CW)			// CW mode
			keyer_mode = eeprom.defval.cw_keyer;
		else											// other beacon modes
			keyer_mode = KEYER_BEACON;

		// Measurements & basic calculations
		power_measurements();


	//*********************************************************************************************
	// ----------> Do front panel switches
	//*********************************************************************************************

		//-----------------------------------------------------------------------------------------
		// DISPLAY / CONFIG
		//-----------------------------------------------------------------------------------------
		if (DISP == 0)													// SW1
		{
			counter.push = LONG_PUSH;									// set long push timer

			if (lcd_mode == 0 && sub_page0 == MAX_SUB_PAGE0)
				beep(TONE_C, cal.calval.beep_len);						// C tone for basic display
			else
			{
				if ((lcd_mode == 1 && sub_page1 == MAX_SUB_PAGE1) || (lcd_mode == 2 && sub_page2 == MAX_SUB_PAGE2))
					beep(TONE_C, cal.calval.beep_len);					// C tone for basic display
				else
					beep(TONE_D, cal.calval.beep_len);					// D tone beep for other displays
			}

			while(DISP == 0)
				if (!counter.push && !eeprom.defval.beacon_tx_on && !eeprom.defval.script_on)
				{
					beep(TONE_C, cal.calval.beep_len*10);				// long beep for major page change
					break;
				}

			if (!counter.push && !eeprom.defval.beacon_tx_on && !eeprom.defval.script_on)	// major mode change
			{
				lcd_mode++;
				if (lcd_mode > MAX_LCD_MODE)
				{
					lcd_mode = 0;
					lcd_cursor_on(1, 1, 0);
					check_wspr_call();
					check_wspr_grid();
					//save_defaults();
					if (beacon_save)save_beacon();
					if (script_save)save_script();
				}
				while(DISP == 0) {}										// wait until button released
			}
			else
			{
				switch (lcd_mode)										// select sub page to adjust
				{
					case 0:
						sub_page0++;
						if (sub_page0 == 4 && eeprom.defval.mode != MODE_WSPR)	// timer only for WSPR mode
							sub_page0++;
						if (sub_page0 > MAX_SUB_PAGE0)
							sub_page0 = 0;
					break;

					case 1:
						sub_page1++;
						if (sub_page1 == 1 && eeprom.defval.mode != MODE_WSPR)	// timer only for WSPR mode
							sub_page1++;
						if (sub_page1 > MAX_SUB_PAGE1)
							sub_page1 = 0;
					break;

					case 2:
						sub_page2++;
						if (sub_page2 > MAX_SUB_PAGE2)
							sub_page2 = 0;
						msg_idx = 0;
						cur_flg = 0;
						disp_idx = 0;
					break;
				}
			}
			encoder_get();												// clear encoder after display change
			ms_delay(BUTTON_DEBOUNCE);
		}


		//-----------------------------------------------------------------------------------------
		// OPER
		//-----------------------------------------------------------------------------------------
		// operating mode select if normal display, long push = beacon TX control
		if (OPER == 0 && lcd_mode == 0)
		{
			counter.push = LONG_PUSH;								// set long push timer

			if (alarms)
				set_pa_state(0, 1, 1, 1);							// if active alarms, always change to standby state
			else
			{
				while(OPER == 0 && eeprom.defval.pa_state == 1)		// long push in OPER mode only
					if (counter.push == 0)
						break;

				if (counter.push == 0 && keyer_mode == KEYER_BEACON)// major mode change
				{
					if (eeprom.defval.mode == MODE_SCRIPT && !eeprom.defval.script_on)		// script was off
						set_script_on_off(1, 1);
					else if (eeprom.defval.mode == MODE_SCRIPT && eeprom.defval.script_on)	// script was on
						set_script_on_off(0, 1);
					else if (eeprom.defval.beacon_tx_on == 0)		// beacon was off
						set_beacon_on_off(1, 1);
					else											// beacon was on
					{
						set_beacon_on_off(0, 1);
						if (eeprom.defval.script_on)
							set_script_on_off(0, 1);
					}
				}
				else
					if (!eeprom.defval.beacon_tx_on && !eeprom.defval.script_on)       // change state if beacon and script are off
						set_pa_state(1, 0, 1, 1);
					
			}
			ms_delay(BUTTON_DEBOUNCE);
			while(OPER == 0) {}										// wait until button released
			clear_alarms();											// try to clear alarms
		}
		else if (OPER == 0 && lcd_mode == 2)
		{
			switch (sub_page2)										// delete character
			{
				case 0:
					key_cw_beacon(0, 3, 1);
				break;

				case 1:
					key_short_msg(0, 3, 1);
				break;

				case 2:
					key_wspr_call(0, 3, 1);
				break;

				case 3:
					key_wspr_grid(0, 3, 1);
				break;

				case 7:
					key_script(0, 3, 1);
				break;
			}
			ms_delay(BUTTON_DEBOUNCE);
			while(OPER == 0) {}
		}

		//-----------------------------------------------------------------------------------------
		// FREQ +
		//-----------------------------------------------------------------------------------------
		if (!FREQ_UP && !lcd_mode && !eeprom.defval.beacon_tx_on && !eeprom.defval.script_on)
		{
			if(RFPWR == 0)
				set_frequency(100, 0, 1, 1);						// speed tune
			else
				set_frequency(1, 0, 1, 1);							// normal tune
			ms_delay(BUTTON_DEBOUNCE);
		}
		else if (FREQ_UP == 0 && lcd_mode == 2)						// adjust character
		{
			switch (sub_page2)
       		{
				case 0:
					key_cw_beacon(1, 1,  0);
				break;

				case 1:
					key_short_msg(1, 1,  0);
				break;

				case 2:
					key_wspr_call(1, 1,  0);
				break;

				case 3:
					key_wspr_grid(1, 1,  0);
				break;

				case 7:
					key_script(1, 1,  0);
				break;
			}
            ms_delay(BUTTON_DEBOUNCE);
		}


		//-----------------------------------------------------------------------------------------
		// FREQ +
		//-----------------------------------------------------------------------------------------
		if (!FREQ_DN && !lcd_mode && !eeprom.defval.beacon_tx_on && !eeprom.defval.script_on)
		{
			if(RFPWR == 0)
				set_frequency(-100, 0, 1, 1);						// speed tune
			else
				set_frequency(-1, 0, 1, 1);							// normal tune
			ms_delay(BUTTON_DEBOUNCE);
		}
		else if (FREQ_DN == 0 && lcd_mode == 2)						// adjust character
		{
			switch (sub_page2)	
			{
				case 0:
					key_cw_beacon(-1, 1, 0);
				break;

				case 1:
					key_short_msg(-1, 1, 0);
				break;

				case 2:
					key_wspr_call(-1, 1, 0);
				break;

				case 3:
					key_wspr_grid(-1, 1, 0);
				break;

				case 7:
					key_script(-1, 1, 0);
				break;
			}
			ms_delay(BUTTON_DEBOUNCE);
		}


		//-----------------------------------------------------------------------------------------
		// RF PWR button alone
		//-----------------------------------------------------------------------------------------
		if (!RFPWR && !lcd_mode && FREQ_UP && FREQ_DN && !eeprom.defval.beacon_tx_on && !eeprom.defval.script_on) 
		{
			set_tx_power(1, 0, 1, 1);
			ms_delay(BUTTON_DEBOUNCE);
			while(RFPWR == 0) {}
		}
		else if (RFPWR == 0 && lcd_mode == 1 && sub_page1 == 6)
		{
			set_tx_power(1, 0, 1, 1);
			ms_delay(BUTTON_DEBOUNCE);
			while(RFPWR == 0) {}
		}
		else if (RFPWR == 0 && lcd_mode == 2)
		{
			switch (sub_page2)										// add character
			{
				case 0:
					key_cw_beacon(0, 2, 1);
				break;

				case 1:
					key_short_msg(0, 2, 1);
				break;
	
				case 2:
					key_wspr_call(0, 2, 1);
				break;

				case 3:
					key_wspr_grid(0, 2, 1);
				break;

				case 7:
					key_script(0, 2, 1);
				break;
			}
			ms_delay(BUTTON_DEBOUNCE);
			while(RFPWR == 0) {}
		}


		//-----------------------------------------------------------------------------------------
		// POWER button
		//-----------------------------------------------------------------------------------------
		if(PWR_SW == 1)
		{
			switch (lcd_mode)
			{
				case 0:
					sub_mode++;
					beacon_lcd_idx = 0;
					switch (eeprom.defval.mode)									// sub mode menu
					{
						case MODE_CW:
							if (sub_mode == 1 && eeprom.defval.cw_keyer != 4)	// no beacon mode
								sub_mode = 0;
							else if (sub_mode == 3 && !eeprom.defval.beacon_tx_on)
								sub_mode = 0;
							else if (sub_mode == 4)
								sub_mode = 0;
						break;

						case MODE_QRSS: case MODE_JASON: case MODE_OPERA: case MODE_WSPR:
							if (sub_mode == 3 && !eeprom.defval.beacon_tx_on)
								sub_mode = 0;
							else if (sub_mode == 4)
								sub_mode = 0;
						break;

						case MODE_DFCW:
							if (sub_mode == 4 && !eeprom.defval.beacon_tx_on)
								sub_mode = 0;
							else if (sub_mode == 5)
								sub_mode = 0;
						break;

						case MODE_SCRIPT:
							if (sub_mode == 2 && !counter.script)
								sub_mode = 0;
							else if (sub_mode == 3)
								sub_mode = 0;
						break;
					}
				break;

				case 1:
					lcd_mode = 2;
					switch (eeprom.defval.mode)
					{
						case MODE_CW:
							sub_page2 = 0;
						break;

						case MODE_QRSS: case MODE_DFCW: case MODE_JASON:
							sub_page2 = 1;
						break;

						case MODE_OPERA: case MODE_WSPR:
							sub_page2 = 2;
						break;

						case MODE_SCRIPT:
							sub_page2 = 7;
						break;
					}
					msg_idx = 0;
					cur_flg = 0;
					disp_idx = 0;
				break;

				case 2:
					lcd_mode = 1;
					lcd_cursor_on(1, 1, 0);
				break;
			}
			power_off();											// go shut down
		}


	//*********************************************************************************************
	// ---------> Beacon
	//*********************************************************************************************
		
		if (eeprom.defval.beacon_tx_on)
			run_beacon();

	//*********************************************************************************************
	// ----------> Script
	//*********************************************************************************************

		if (eeprom.defval.script_on)
			run_script();

	//*********************************************************************************************
	// ----------> Do LCD display and UP/DOWN buttons
	//*********************************************************************************************

		switch (lcd_mode)
		{
			//-------------------------------------------------------------------------------------
			// Normal display
			//
			// |A: mode parameters 8 char     |B: meter 7 char       |
			// |C:mode 4 char|D:pa state 6 char  |E: frequency 6 char|
			//-------------------------------------------------------------------------------------
			case 0: default:

				//-----> A
				// select beacon message to display
				if (keyer_mode != KEYER_BEACON)					// not in beacon mode
				{
					lcd_cursor(1, 1);
					lcd_s_meter(pwr_meter);						// show PWR meter bar graph
					lcd_putch(S_BAR);							// PWR-meter stopper bar
				}
				else
				{
					if(eeprom.defval.beacon_tx_on == 0)			// action only if beacon off
						temp = encoder_get();
					else
						temp = 0;
					switch (sub_mode)
					{
						case 0:
							switch (eeprom.defval.mode)
							{
								case MODE_CW:
									set_cw_speed(temp, 0, 1, 1);
								break;

								case MODE_QRSS: case MODE_DFCW:
									set_dot_time(temp, 0, 1, 1);
								break;

								case MODE_JASON:
									set_jason_speed(temp, 0, 1, 1);
								break;

								case MODE_OPERA:
									set_opera_speed(temp, 0, 1, 1);
								break;

								case MODE_WSPR:
									set_wspr_speed(temp, 0, 1, 1);
								break;

								case MODE_SCRIPT:
									set_script_frame(temp, 0, 1, 1);
								break;
							}
						break;

						case 1:
							switch (eeprom.defval.mode)
							{
								case MODE_CW: case MODE_QRSS:
									set_cw_frame(temp, 0, 1, 1);
								break;

								case MODE_DFCW:
									set_cw_shift(temp, 0, 1, 1);
								break;

								case MODE_JASON:
									set_jason_frame(temp, 0, 1, 1);
								break;

								case MODE_OPERA:
									set_opera_frame(temp, 0, 1, 1);
								break;

								case MODE_WSPR:
									set_wspr_frame(temp, 0, 1, 1);
								break;
								
								case MODE_SCRIPT:
									disp_beacon_lcd(1);
								break;
							}
						break;

						case 2:
							if (eeprom.defval.mode == MODE_DFCW)
								set_cw_frame(temp, 0, 1, 1);
							else if (eeprom.defval.mode == MODE_SCRIPT)
								set_script_delay(0,1);
							else
								disp_beacon_lcd(1);
						break;

						case 3:
							if (eeprom.defval.mode == MODE_DFCW)
								disp_beacon_lcd(1);
							else
								disp_beacon_in_progress();
						break;

						case 4:
							if (eeprom.defval.mode == MODE_DFCW || eeprom.defval.mode == MODE_CW)	//CW for CWID
								disp_beacon_in_progress();
						break;
					}	
				}
			
				//-----> B
				// select meter to display
				switch (sub_page0)								// select meter
				{
					case 0: default:							// Power
						disp_fwd_pwr();
					break;

					case 1:										// SWR forward power
						disp_swr();
					break;

					case 2:										// battery voltage mean value, measure only when display selected
						disp_batt_volt();
					break;

					case 3:										// Drain current, measure only when display selected
						disp_id();
					break;

					case 4:										// WSPR timer
						if (eeprom.defval.mode == MODE_WSPR)
							disp_timer(0);
						else									// mode changed
						{
							sub_page0 = 0;
							disp_fwd_pwr();
						}
					break;
				}

				//-----> C and UP/DOWN
				switch (eeprom.defval.mode)
				{
					case MODE_CW:
						if (eeprom.defval.cw_keyer == 4)		// beacon mode
						{
							if (flag.cwid)
								lcd_display_at(2, 1 , "CWID ");
							else
								lcd_display_at(2, 1 , "CW   ");
						}
						else
						{
							temp = encoder_get();				// keyer modes
							set_cw_speed(temp, 0, 1, 1);
						}
					break;

					case MODE_QRSS:
						lcd_display_at(2, 1 , "QRSS ");
					break;

					case MODE_DFCW:
						lcd_display_at(2, 1 , "DFCW ");
					break;

					case MODE_JASON:
						lcd_display_at(2, 1 , "JSON ");
					break;

					case MODE_OPERA:
						lcd_display_at(2, 1 , "OPRA ");
					break;

					case MODE_WSPR:
						lcd_display_at(2, 1 , "WSPR ");
					break;

					case MODE_SCRIPT:
						lcd_display_at(2, 1 , "SCRI ");
					break;
				}

				//-----> D
				// display PA state
				if (alarms)
					disp_alarms();
				else
				{
					set_pa_state(0, 0, 0, 1);
					if (eeprom.defval.script_on)				// display warning during script
						lcd_display_at(2, 10 , "!");
					else
						lcd_display_at(2, 10 , " ");
				}

				//-----> E
				// display frequency
				set_frequency(0, 0, 0, 1);
			break;

			//-------------------------------------------------------------------------------------
			// User config display
			//
			// |Configuration type line 16 char                      |
			// |Configuration data Line 16 char                      |
			//-------------------------------------------------------------------------------------

			//-----> type & data line and UP/DOWN
			// select configuration to display and adjust if UP / DOWN buttons pressed
			case 1:
				switch(sub_page1)
				{
					case 0:
						if (set_mode(encoder_get(), 0, 1, 1))
							sub_mode = 0;
					break;

					case 1:
						set_wspr_timer(encoder_get(), 0, 0, 0, 1, 1);
					break;

					case 2:
						set_preamp(encoder_get(), 0, 1, 1);
					break;

					case 3:
						set_converter(encoder_get(), 0, 1, 1);
					break;

					case 4:
						set_cw_keyer(encoder_get(), 0, 1, 1);
					break;

					case 5:
						set_cw_sidetone(encoder_get(), 0, 1, 1);
					break;

					case 6:
						set_swr_limit(encoder_get(), 0, 1, 1);
					break;

					case 7: default:
						set_back_light(encoder_get() * 10, 0, 1, 1);
					break;

					case 8:
						set_contrast(encoder_get(), 0, 1, 1);
					break;

					case 9:
						set_serial_mode(encoder_get(), 0, 1, 1);
					break;

					case 10:
						set_baud_rate(encoder_get(), 0, 1, 1);
					break;

					case 11:
						set_mox(encoder_get(), 0, 1, 1);
					break;

					case 12:
						set_spare_io(encoder_get(), 0, 1, 1);
					break;
				}
			break;

			//-------------------------------------------------------------------------------------
			// Beacon setup display
			//
			// |Beacon setup type line 16 char                       |
			// |Beacon setup data Line 16 char                       |
			//-------------------------------------------------------------------------------------

			//-----> type & data line and UP/DOWN
			// select beacon message to display and move cursor or adjust if UP / DOWN buttons pressed
            case 2:
                switch (sub_page2)
                {
                    case 0:
						key_cw_beacon(encoder_get(), 0,  0);
					break;

					case 1:
						key_short_msg(encoder_get(), 0,  0);
					break;

					case 2:
						key_wspr_call(encoder_get(), 0,  0);
					break;

					case 3:
						key_wspr_grid(encoder_get(), 0,  0);
					break;

					case 4:
						set_wspr_ndbm(encoder_get() * 3, 0, 1, 1);
					break;

					case 5:
						set_wspr_gps(encoder_get(),0, 1, 1);
					break;

					case 6:
						set_cwid(encoder_get(), 0,  1, 1);
					break;

					case 7:
						key_script(encoder_get(),0, 0);
					break;
				}
			break;
		}

	//*********************************************************************************************
	// ----------> Do serial interface
	//*********************************************************************************************
		switch(eeprom.defval.serial_mode)
		{
			case 0:
				if (!eeprom.defval.script_on)	// diseable if script in progress
					serial_tx500(0);			// TX500 / TX136 protocol
			break;

			case 1: default:
				serial_test();					// test commands
			break;

			// UART1 can be used for GPS if UART2 is not wired
			case 2:
				if (eeprom.defval.mode == MODE_WSPR)
					serial_gps(0);				// GPS NMEA protocol
			break;
		}

		// UART2 used for GPS by default - UART2 must be wired
		if (eeprom.defval.serial_mode != 2 && eeprom.defval.mode == MODE_WSPR)
			serial_gps(1);						// GPS NMEA protocol

	}
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Service & calibration functions, when tested move to own module
//=================================================================================================
void service(void)
{
	int cal_page;

	// display setup
	cal_page = 0;
	lcd_clear();

	// conditional start up text printout, print only if RS232 = Test
 	if(eeprom.defval.serial_mode == 1)
	{
		printf("\n\rService & calibration mode");
		printf("\n\rDISP = parameter select, OPER = save & exit");
	}

//*************************************************************************************************
// ----------> Service forever loop
//*************************************************************************************************
	for(;;)
	{
		// Make atomic copy of KEY(TX request) signal for this processing round
		if(PTT_OUT_1)						// 1 = active
			key = 1;						// copy I/O bit to status flag	
		else
			key = 0;

		// do measurements & check alarms
		power_measurements();
		check_alarms();


	//*********************************************************************************************
	// ----------> Do front panel switches
	//*********************************************************************************************

		//-----------------------------------------------------------------------------------------
		// DISPLAY / CONFIG
		//-----------------------------------------------------------------------------------------
		if(DISP == 0)							// select calibration function
		{
			cal_page++;
			if(cal_page > MAX_SERVICE_PAGES) 	// number of active pages
				cal_page = 0;
			beep(HZ392_01, 50);
			while(DISP == 0) {}
			ms_delay(BUTTON_DEBOUNCE);
			encoder_get();						// flush encoder
		}

		switch(cal_page)						// select major page to display
		{
			case 0: default:					// local oscillator
				set_cal_oscillator(encoder_get() * 10, 0, 1, 1);
			break;

			case 1:								// battery voltage
				set_cal_batt(encoder_get(), 0, 1, 1);
			break;

			case 2:								// beep length
				set_cal_beep(encoder_get(), 0, 1, 1);
			break;

			case 3:								// Forward power scaling
				set_cal_fwr_pwr(encoder_get(), 0, 1, 1);
			break;

			case 4:								// ID scaling
				set_cal_id(encoder_get(), 0, 1, 1);
			break;

			case 5:								// CW break period
				set_cal_cw_break(encoder_get(), 0, 1, 1);
			break;

			case 6:								// offset for WSPR timer
				set_cal_ms_offset(encoder_get(), 0, 1, 1);
			break;

			case 7:								// jumper Q4 value
				set_cal_power_q4(encoder_get(), 0, 1, 1);
			break;

			case 8:								// factory defaults, this must be last item in the list
				lcd_display_all("Push OPER long =", "factory defaults"); 
			break;
		}


		//-----------------------------------------------------------------------------------------
		// OPER
		//-----------------------------------------------------------------------------------------

		if(OPER == 0 && cal_page == MAX_SERVICE_PAGES)			// Restore defaults only when last cal_page
		{
			counter.push = LONG_PUSH;							// set long push timer
				
			while (OPER == 0)									// detect long push
			{
				if(counter.push == 0)							// detect long push, restore factory setup
				{
					set_factory_defaults();						// EEPROM empty, set factory defaults
					save_calval();								// save calibration values
					save_defaults();							// save user defaults
						
					beep(HZ466_85, cal.calval.beep_len*10);		// long beep for save
					service_mode = 0;							// back to normal operation
					lcd_clear();
					lcd_display_line(1, "Factory setup ok"); 
						
 					if(eeprom.defval.serial_mode == 1)			// conditional print only if RS232 = Test
						printf("\n\rFactory defaults restored");
					ms_delay(1500);								// show message
					while(OPER == 0) {}
					return;										// exit service mode
				}
			}
		}
		else if (OPER == 0)										// test SW2 (SET) when  cal_page != 8
		{
			save_calval();										// save changes
			beep(HZ466_85, cal.calval.beep_len);				// short beep for return to normal operation
			service_mode = 0;									// back to normal operation
			lcd_clear();
			lcd_display_line(1, "Calibr. Saved"); 
		
			if(eeprom.defval.serial_mode == 1)					// conditional print only if RS232 = Test
				printf("\n\rCalibration settings saved to EEPROM");
			while(OPER == 0) {}
			return;												// exit service mode
		}	


		//-----------------------------------------------------------------------------------------
		// POWER
		//-----------------------------------------------------------------------------------------
		if(PWR_SW == 1)
			power_off();			// go shut down, do not save changes

	}
}
//-------------------------------------------------------------------------------------------------
