//*************************************************************************************************
//
// JUMA-TX136/TX500 protocol
//
// Juha Niinikoski, OH2NLT 09.09.2009
//
// Beacon repeat count logic added 14.09.2009
// very slow CW keyer speed, downto 0.1 wpm, S command changed to 3-digits 20.09.2009
// beacon message '=' and '?' character behaviour corrected 17.10.2009
// Add QRSS and DFCW modes - F4GCB 08.2014
// Add WSPR mode - F4GCB 10.2014
// Program structure modified to improve readability and maintenance - F4GCB 02-2015
// Add UART2 for optional GPS NMEA - 05.2015
// Add OPERA mode - F4GCB 05.2015
// Add JASON mode - F4GCB 08.2015
// Add CW identity option - F4GCB 04.2016
// Bug fixed: character '=' accepted in QRSS/DFCW/JASON beacon message - F4GCB 04.2016
// Add SCRIPT mode - F4GCB 04.2016
//
//*************************************************************************************************

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "juma-tx500.h"
#include "eeprom_tx500.h"
#include "timers_pwm.h"
#include "uart.h"
#include "uart2.h"


// Local definitions ------------------------------------------------------------------------------
#define CMD_BUF_SIZE 20							// for full size command message
#define MSG_GPS_SIZE 75							// for GPS NMEA message

// Local variables --------------------------------------------------------------------------------
static char cmd_buf[BEACON_BUFFER_SIZE];		// command buffer										/
static int cmd_buf_idx = 0;						// buffer fill pointer									|
static char bcn_buf[BEACON_BUFFER_SIZE-2];		// beacon message buffer								|
static int bcn_buf_idx = 0;						// beacon message buffer fill index						|> command serial buffer
static int rd_bcn_msg_flag = 0;					// instead of command interpreter read beacon message   |
static char s_buf[SCRIPT_BUFFER_SIZE];			// script buffer								|
static int s_buf_idx = 0;						// script buffer fill index						|
static char *s_buf_ptr;							// script buffer current pointer				/
static int gps_status = 0;						// GPS status 0=nok				/
static char msg_gps[MSG_GPS_SIZE];				// GPS message buffer			|
static int msg_gps_idx = 0;						// GPS buffer fill pointer		|> GPS serial buffer
static char grid4[5];							// GPS grid 4					|
static char grid6[7];							// GPS grid 6					/


//*************************************************************************************************
// TX136/500 serial beacon set functions
//*************************************************************************************************

//=================================================================================================
// Collect message into the beacon message buffer, CR terminate the operation
//=================================================================================================
void collect_beacon_message(char c)
{
	bcn_buf[bcn_buf_idx] = c;

	if(bcn_buf_idx < BEACON_BUFFER_SIZE - 1)
		bcn_buf_idx++;
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Collect message into the script buffer, CR terminate the operation
//=================================================================================================
void collect_script_message(char c)
{
	s_buf[s_buf_idx] = c;

	if(s_buf_idx < SCRIPT_BUFFER_SIZE - 1)
		s_buf_idx++;
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Terminate message collect, we have received CR
//=================================================================================================
void set_beacon_message(void)
{
	//bcn_buf[bcn_buf_idx] = 0x0D;		// write terminator
	//bcn_buf_idx++;
	bcn_buf[bcn_buf_idx] = 0x00;
	set_cw_beacon(bcn_buf, 0, 1);
	init_beacon();						// init beacon logic, need this if beacon running
	rd_bcn_msg_flag = 0;
	cur_flg = 0;
}
//-------------------------------------------------------------------------------------------------


//*************************************************************************************************
// TX136/500 serial commands functions
//*************************************************************************************************

//=================================================================================================
// Convert number from command buffer
//=================================================================================================
static unsigned long get_number(int nb_idx)
{
	unsigned long number = 0;
	int i = nb_idx;			// msb of the number

	while(i < cmd_buf_idx)
	{
		number = (number * 10) + (cmd_buf[i] & 0x0F);
		i++;
	}

	return number;
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Clear buffer
//=================================================================================================
void clear_buffer(void)
{
	int x;

	// clear whole buffer
	for(x=0; x<CMD_BUF_SIZE; x++)
	// clear only CMD bytes
	//for(x=0; x<9; x++)
		cmd_buf[x] = 0;
	cmd_buf_idx = 0;
	rd_bcn_msg_flag = 0;
	// reset also buffer data collect index
	bcn_buf_idx = 0;
	s_buf_idx = 0;
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Filter no operation characters, return = 1 if no op
//=================================================================================================
int filter_nop_char(char c)
{
	//if((c==0x00) || (c==0x0A) || (c==0x20) || (c==0x2C) || (c==0x2E))
	// we need space and comma and period for beacon string
	if((c==0x00) || (c==0x0A))
		return 1;
	else
		return 0;
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Parse TX500/TX136 message or script
//=================================================================================================
void serial_tx500(int script)
{
	char c;
	char buffer[250];

	// process one character at time
	//if(kbhit())
	// process all at once, better for high speed
	while((!script && uart1_kbhit()) || script)
	{
		if (script)
		{
			c = 0x0D;						// put the buffer ready
			script = 0;
		}
		else
			c = uart1_get_char();			// save character

		// discard nop characters
		if(filter_nop_char(c))
			return;

		// capture start sync with action characters, when collecting beacon buffer accept them
		if(c == '=' || c == '?')
			// normal mode & command received
			if(rd_bcn_msg_flag == 0 && cmd_buf_idx > 2 )
				clear_buffer();

		// check if buffer ready, interpret message - CR = message ready, action
		if(c == 0x0D)
		{
			switch(cmd_buf[0])
			{
				// send values
				case '?':
					// select command
					switch(cmd_buf[1])
					{
						case 'A':
							printf("=A%i\n\r", eeprom.defval.preamp);
						break;
	
						case 'B':
							printf("=B%i\n\r", eeprom.defval.beacon_tx_on);
						break;
	
						case 'C':
							printf("=C%i\n\r", eeprom.defval.converter);
						break;

						case 'D':
							printf("=D%.3d\n\r", eeprom.defval.cw_dot_time);
						break;

						case 'F':
							printf("=F%.6ld\n\r", eeprom.defval.txfreq[board_type]);
						break;

						case 'G':
							printf("=G%i\n\r", eeprom.defval.mode);
						break;

						case 'H':
							// select sub command
							switch(cmd_buf[2])
							{
								case 'S':default:
									printf("=HS%.1d\n\r", eeprom.defval.opera_speed);
								break;

								case 'F':
									printf("=HF%.1d\n\r", eeprom.defval.opera_frame);
								break;
							}
						break;

						case 'I':
							// select sub command
							switch(cmd_buf[2])
							{
								case 'B':
									printf("=IB%ld\n\r", meas_batt_volt());
								break;
	
								case 'D':
									printf("=ID%ld\n\r", meas_id());
								break;

								case 'I': default:
									printf("=II%s\n\r", get_info(buffer));
								break;

								case 'P':
									printf("=IP%d\n\r", get_out_pwr());
								break;

								case 'S':
									printf("=IS%d\n\r", get_swr());
								break;
							}
						break;

						case 'J':
							// select sub command
							switch(cmd_buf[2])
							{
								case 'S':default:
									printf("=JS%i\n\r", eeprom.defval.jason_speed);
								break;

								case 'F':
									printf("=JF%.2d\n\r", eeprom.defval.jason_frame);
								break;
								}
						break;

						case 'K':
							printf("=K%i\n\r", eeprom.defval.cw_keyer);
						break;

						case 'L':
							printf("=L%s\n\r", eeprom.defval.short_msg);
						break;

						case 'M':
							printf("=M%s\n\r", beacon_buffer);
						break;

						case 'O':
							printf("=O%i\n\r", eeprom.defval.pa_state);
						break;

						case 'P':
							printf("=P%i\n\r", eeprom.defval.rfpwr);
						break;

						case 'Q':
							printf("=Q%.2d\n\r", eeprom.defval.cw_frame);
						break;

						case 'R':
							printf("=R%.2d\n\r", eeprom.defval.cw_shift);
						break;

						case 'S':
							printf("=S%.2d\n\r", eeprom.defval.cw_speed);
						break;

						case 'T':
							printf("=T%i\n\r", eeprom.defval.mox);
						break;

						case 'V':
							printf("=V%s\n\r", script_buffer);
						break;

						case 'W':
							// select sub command
							switch(cmd_buf[2])
							{
								case 'C':
									printf("=WC%s\n\r", eeprom.defval.wspr_call);
								break;
	
								case 'F':
									printf("=WF%.1d\n\r", eeprom.defval.wspr_frame);
								break;

								case 'G':
									printf("=WG%.1d\n\r", eeprom.defval.wspr_gps);
								break;

								case 'L':
									printf("=WL%s\n\r", eeprom.defval.wspr_grid);
								break;

								case 'P':
									printf("=WP%.2d\n\r", eeprom.defval.wspr_ndbm);
								break;

								case 'R':
									printf("=WR%s\n\r", grid6);
								break;

								case 'S':default:
									printf("=WS%.1d\n\r", eeprom.defval.wspr_speed);
								break;

								case 'T':
									printf("=WT%.4d\n\r", counter.wspr);
								break;
							}
						break;

						case 'X':
							printf("=X%i\n\r", eeprom.defval.spare_io);
						break;

						case 'Y':
							printf("=Y%i\n\r", eeprom.defval.cwid);
						break;

						case 'Z':
						// select sub command
						switch(cmd_buf[2])
						{
							case 'F':
								printf("=ZF%.1d\n\r", eeprom.defval.script_frame);
							break;
						}
						break;
					}
					// clear for next round
					clear_buffer();
				break;

				// set values
				case '=':
					// select command
					switch(cmd_buf[1])
					{
						case 'A':
							set_preamp(get_number(2), 1, 0, 0);
						break;

						case 'B':
							if (eeprom.defval.mode == MODE_SCRIPT)
								set_script_on_off(get_number(2), 1);
							else
								set_beacon_on_off(get_number(2), 1);
						break;

						case 'C':
							set_converter(get_number(2), 1, 0, 0);
						break;

						case 'D':
							set_dot_time(get_number(2), 1, 0, 0);
						break;

						//case 'E':				// do by default with 'M' command
						//	save_beacon();
						//break;

						case 'F':
							set_frequency(get_number(2), 1, 0, 1);
						break;

						case 'G':
							set_mode(get_number(2), 1, 0, 0);
						break;

						case 'H':
							// select sub command
							switch(cmd_buf[2])
							{
								case 'S':
									set_opera_speed(get_number(3), 1, 0, 0);
								break;

								case 'F':
									set_opera_frame(get_number(3), 1, 0, 0);
								break;
							}
						break;

						case 'J':
							// select sub command
							switch(cmd_buf[2])
							{
								case 'S':
									set_jason_speed(get_number(3), 1, 0, 0);
								break;

								case 'F':
									set_jason_frame(get_number(3), 1, 0, 0);
								break;
							}
						break;

						case 'K':
							set_cw_keyer(get_number(2), 1, 0, 0);
						break;

						case 'L':
							set_short_msg(cmd_buf + (2 * sizeof(char)), 1);
							cur_flg = 0;
						break;

						case 'M':
							set_beacon_message();
							save_beacon();
						break;

						case 'O':
							set_pa_state(get_number(2), 1, 0, 0);
						break;

						case 'P':
							set_tx_power(get_number(2), 1, 0, 0);
						break;

						case 'Q':
							set_cw_frame(get_number(2), 1, 0, 0);
						break;

						case 'R':
							set_cw_shift(get_number(2), 1, 0, 0);
						break;

						case 'S':
							set_cw_speed(get_number(2), 1, 0, 0);
						break;

						case 'T':
							set_ptt_out(get_number(2));
						break;

						case 'V':
								s_buf[s_buf_idx] = 0x00;
								set_script(s_buf, 0);
								cur_flg = 0;
								save_script(); 
						break;

						case 'W':
							// select sub command
							switch(cmd_buf[2])
							{
								case 'C':
									set_wspr_call(cmd_buf + (3 * sizeof(char)), 1);
									cur_flg = 0;
								break;

								case 'F':
									set_wspr_frame(get_number(3), 1, 0, 0);
								break;

								case 'G':
									set_wspr_gps(get_number(3), 1, 0, 0);
								break;

								case 'L':
									set_wspr_grid(cmd_buf + (3 * sizeof(char)), 1);
									cur_flg = 0;
								break;

								case 'P':
									set_wspr_ndbm(get_number(3), 1, 0, 0);
								break;

								case 'S':
									set_wspr_speed(get_number(3), 1, 0, 0);
								break;

								case 'T':
									if (counter.gps == 0)
										set_wspr_timer(get_number(3), 1, 0, 0, 0, 0);
								break;
							}
						break;

						case 'X':
							set_spare_io(get_number(2), 1, 0, 0);
						break;

						case 'Y':
							set_cwid(get_number(2), 1, 0, 0);
						break;

						case 'Z':
						// select sub command
							switch(cmd_buf[2])
							{
								case 'F':
									set_script_frame(get_number(3), 1, 0, 0);
								break;

								case 'T':
									set_script_delay(get_number(3), 0);
								break;
							}
						break;
					}
					// clear for next round
					clear_buffer();
				break;

				// not known
				default:
					// clear for next round
					clear_buffer();
				break;
			}
		}

		else
		{
			// special case ('=' as character message), collect to the message buffer
			if(cmd_buf[0] == '=' && cmd_buf[1] == 'L')
				rd_bcn_msg_flag = 1;
			if(cmd_buf[0] == '=' && cmd_buf[1] == 'M')
			{
				rd_bcn_msg_flag = 1;
				collect_beacon_message(c);
			}
			else if(cmd_buf[0] == '=' && cmd_buf[1] == 'V')
			{
				rd_bcn_msg_flag = 1;
				collect_script_message(c);
			}
			else
			{
				cmd_buf[cmd_buf_idx] = c;
				if(cmd_buf_idx < CMD_BUF_SIZE-1)
					cmd_buf_idx++;
			}
		}
	}
}
//-------------------------------------------------------------------------------------------------


//*************************************************************************************************
// Script functions
//*************************************************************************************************

//=================================================================================================
// Init script
//=================================================================================================
void serial_script_init(const char *str)
{
	strcpy(s_buf, str);
	s_buf_ptr = s_buf;					// set current pointer
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// This function is called continuously from main when script is on
// init must be done before
// return 1 when script is finished
//=================================================================================================
int serial_script_run(void)
{
	char c;
	
	clear_buffer();
	do
	{
		// get character
		c = *s_buf_ptr;

		// test of the command sequence
		if (((c == '=' || c == '?') || (c == 0 || c == 0x0D))&& cmd_buf_idx)
		{
			serial_tx500(1);					// play the script
			clear_buffer();
			return 0;
		}
		// test end of script
		else if (c == 0 || c == 0x0D)
		{
			// start over
			s_buf_ptr = s_buf;
			return 1;
		}
		// get command character
		else
		{
			cmd_buf[cmd_buf_idx] = c;
			if(cmd_buf_idx < CMD_BUF_SIZE-1)
				cmd_buf_idx++;
			s_buf_ptr++;
		}
	}
	while(1);

	return 0;
}
//-------------------------------------------------------------------------------------------------


//*************************************************************************************************
// GPS functions
//*************************************************************************************************

//=================================================================================================
// Clear buffer GPS
//=================================================================================================
void clear_gps(void)
{
	int x;

	// clear whole buffer
	//for(x=0; x<MSG_GPS_SIZE; x++)
	// clear only the sentence type bytes
	for(x=0; x<6; x++)
		msg_gps[x] = 0;
	msg_gps_idx = 0;
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Define locator in function of longitude and latitude
//=================================================================================================
void serial_grid (float longitude, float latitude, char *grid)
{
	float symb1, symb2, symb3, symb4, symb5, symb6;

	symb1 = ((180 + longitude) / 20);
    symb2 = ((90 + latitude) / 10);
    symb3 = (symb1 - (int)symb1) * 10;
    symb4 = (symb2 - (int)symb2) * 10;
    symb5 = (symb3 - (int)symb3) * 24;
  	symb6 = (symb4 - (int)symb4) * 24;

    grid[0] = (int)symb1 + 65;
    grid[1] = (int)symb2 + 65;
    grid[2] = (int)symb3 + 48;
    grid[3] = (int)symb4 + 48;
    grid[4] = (int)symb5 + 65;
    grid[5] = (int)symb6 + 65;
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Parse GPS NMEA message - UART 1 = 0, UART 2 = 1
//=================================================================================================
void serial_gps(int uart)
{
	char c;
	const char gps_start[] = {"$GPGGA"}; 			// starting GPS message
	int i, idx, cnt, minutes, seconds, millis;
	int sep[15];
	float longitude, latitude;

	// process all at once, better for high speed
	while((!uart && uart1_kbhit()) || (uart && uart2_kbhit()))
	{
		if (uart)
			c = uart2_get_char();
		else
			c = uart1_get_char();

		// check if buffer ready, interpret message - LF = message ready, action
		if(c == 0x0A)
		{
      		cnt = 0;

			// check the GPS
			if (msg_gps[0] == '$')
				gps_status = 1;
			else
				gps_status = 0;

			// werify the received message starts with $GPGGA
      		for (i=0; i<6; i++)
        		if (msg_gps[i] == gps_start[i])
          			cnt++;

			// message awaited
			if (cnt == 6)
			{
				idx = 0;
				for (i=6; i<msg_gps_idx; i++)
				{
					// check the position separators
					if (msg_gps[i] == ',')
					{
						sep[idx] = i;
						idx++;
					}
				}

				// valid datas
				if (idx == 14 && msg_gps[sep[5]+1] != '0')
				{
					// disable TMR3 tick irq
					_T3IE = 0;

					// define WSPR timer
        			//hour = (msg_gps[sep[0]+1] - 48) * 10 + msg_gps[sep[0]+2] - 48;
        			minutes = (msg_gps[sep[0]+3] - 48) * 10 + msg_gps[sep[0]+4] - 48;
        			seconds = (msg_gps[sep[0]+5] - 48) * 10 + msg_gps[sep[0]+6] - 48;
					millis = (msg_gps[sep[0]+8] - 48) * 100 + (msg_gps[sep[0]+9] - 48) * 10;

					// some GPS with hundredth only
					if (sep[1] > sep[0]+10)
						millis = millis + (msg_gps[sep[0]+10] - 48);

					// set WSPR timer
					set_wspr_timer((minutes * 60) + seconds, 0, 1, millis, 0, 0);

					// enable TMR3 tick irq
					_T3IE = 1;

					// define locator
					longitude = (msg_gps[sep[3]+1] - 48) * 100 + (msg_gps[sep[3]+2] - 48) * 10 + msg_gps[sep[3]+3] - 48
								+ (msg_gps[sep[3]+4] - 48) * 0.1 + (msg_gps[sep[3]+5] - 48) * 0.01;
					if (msg_gps[sep[4]+1] == 'W')
						longitude = -longitude;
					latitude = (msg_gps[sep[1]+1] - 48) * 10 + msg_gps[sep[1]+2] - 48
							 + (msg_gps[sep[1]+3] - 48) * 0.1 + (msg_gps[sep[1]+4] - 48) *0.01;
					if (msg_gps[sep[2]+1] == 'S')
						latitude = -latitude;
					serial_grid(longitude, latitude, grid6);
					strncpy(grid4, grid6, 4);

					// datas not used
					/*satellite = (msg_buf[sep[6]+1] - 48) * 10 + msg_buf[sep[6]+2] - 48;*/
				}
			}

			// clear for next round
			clear_gps();
		}

		// collect data to the command buffer
		else
		{
			msg_gps[msg_gps_idx] = c;
			if(msg_gps_idx < MSG_GPS_SIZE-1)
				msg_gps_idx++;
		}
	}

	// switch baud rate if no GPS message
	if (!gps_status)
		uart2_set_baud();
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Get GPS grid 4
//=================================================================================================
char *get_gps_grid4(char *string)
{
	return strcpy(string, grid4);
}
//-------------------------------------------------------------------------------------------------


//=================================================================================================
// Get GPS grid 6
//=================================================================================================
char *get_gps_grid6(char *string)
{
	return strcpy(string, grid6);
}
//-------------------------------------------------------------------------------------------------
